const std = @import("std");

const out = std.io.getStdOut().writer();

const WORKER_SIZE = 32;

const READ_SIZE = 2048 * 2048; //2048 * 2048;

const InfoToKeepTrack = struct {
    min: i32,
    max: i32,
    mean: i64,
    count: usize,
};

const DataHash = std.StringArrayHashMap(*InfoToKeepTrack);

fn parseFloat(num: []const u8) i32 {
    const mult: i32 = if (num[0] == '-') -1 else 1;
    const n = num[if (num[0] == '-') 1 else 0..];

    // x.x
    if (n.len == 3) {
        var v = @Vector(2, i32){ n[0], n[2] };
        const v2 = @Vector(2, i32){ '0', '0' };

        v = v - v2;
        return mult * (v[0] * 10 + v[1]);
        // xx.x
    } else if (n.len == 4) {
        const v = @Vector(3, i32){ @intCast(n[0]), @intCast(n[1]), @intCast(n[3]) };
        const v2 = @Vector(3, i32){ '0', '0', '0' };
        const v3 = @Vector(3, i32){ 100, 10, 1 };

        const nv = (v - v2) * v3;
        return mult * @as(i32, @reduce(.Add, nv));
    } else {
        unreachable;
    }
}

fn processLines(alloc: std.mem.Allocator, hashmap: *DataHash, buf: []u8) void {
    var iter = std.mem.splitScalar(u8, buf, '\n');
    while (iter.next()) |line| {
        if (line.len == 0) {
            continue;
        }

        const index = std.mem.lastIndexOfScalar(u8, line, ';').?;

        const name = line[0..index];
        const number = parseFloat(line[index + 1 ..]);

        if (hashmap.get(name)) |v| {
            var value = v;
            value.count += 1;
            value.max = @max(value.max, number);
            value.min = @min(value.min, number);
            value.mean += @intCast(number);
        } else {
            const new_info = alloc.create(InfoToKeepTrack) catch unreachable;
            new_info.count = 1;
            new_info.max = number;
            new_info.min = number;
            new_info.mean = @intCast(number);
            hashmap.putAssumeCapacity(name, new_info);
        }
    }
}

fn count_fn(ptr: []u8, buf: []u64, finish: *bool, ghash: *DataHash, ghash_mutex: *std.Thread.Mutex, working: *bool) void {
    var gpa = std.heap.GeneralPurposeAllocator(.{ .thread_safe = false }){};
    defer _ = gpa.deinit();

    const alloc = gpa.allocator();

    var internal_hash_map = DataHash.init(alloc);
    defer internal_hash_map.deinit();
    internal_hash_map.ensureTotalCapacity(5000) catch unreachable;

    while (!finish.* or working.*) {
        if (working.*) {
            processLines(alloc, &internal_hash_map, ptr[buf[0]..buf[1]]);
            working.* = false;
        } else {
            std.time.sleep(1);
        }
    }
    ghash_mutex.lock();
    for (internal_hash_map.keys(), internal_hash_map.values()) |k, iv| {
        if (ghash.get(k)) |value| {
            var v = value;
            v.max = @max(v.max, iv.max);
            v.min = @min(v.min, iv.min);
            v.mean = v.mean + iv.mean;
            v.count += iv.count;
        } else {
            ghash.putAssumeCapacity(k, iv);
        }
    }
    ghash_mutex.unlock();
}

fn lessThan(_: void, lhs: []const u8, rhs: []const u8) bool {
    return std.mem.order(u8, lhs, rhs) == .lt;
}

pub fn main() !void {
    var gpa = std.heap.GeneralPurposeAllocator(.{ .thread_safe = false }){};
    defer _ = gpa.deinit();

    const alloc = gpa.allocator();

    var file = try std.fs.cwd().openFile("../measurements.txt", .{});
    defer file.close();

    const md = try file.metadata();
    const size = md.size();

    const ptr = try std.posix.mmap(
        null,
        size,
        std.posix.PROT.READ,
        .{ .TYPE = .SHARED },
        file.handle,
        0,
    );
    defer std.posix.munmap(ptr);

    var buf: [WORKER_SIZE][2]u64 = undefined;

    var threads: [WORKER_SIZE]std.Thread = undefined;
    var threads_mux: [WORKER_SIZE]bool = undefined;

    var finish = false;

    var count = DataHash.init(alloc);
    defer count.deinit();

    try count.ensureTotalCapacity(5000);

    var mutex = std.Thread.Mutex{};

    var round: usize = 0;

    var reading_more_faster_than_mux: u64 = 0;

    for (0..WORKER_SIZE) |i| {
        buf[i] = std.mem.zeroes([2]u64);
        threads_mux[i] = false;
    }

    var p: u64 = 0;

    while (true) {
        const end = p + READ_SIZE;
        if (end > size) {
            mutex.lock();
            processLines(alloc, &count, ptr[p..]);
            mutex.unlock();
            break;
        }
        if (!threads_mux[round]) {
            buf[round][0] = p;
            p += std.mem.lastIndexOf(u8, ptr[p..end], "\n").? + 1;
            buf[round][1] = p - 1;
            threads_mux[round] = true;
        } else {
            if (reading_more_faster_than_mux == 0) {
                for (0..WORKER_SIZE) |i| {
                    threads[i] = try std.Thread.spawn(.{}, count_fn, .{ ptr, &buf[i], i, &finish, &count, &mutex, &threads_mux[i] });
                }
            }
            reading_more_faster_than_mux += 1;
        }
        comptime {
            if (WORKER_SIZE != 32) @compileError("ONLY WORKS FOR 32");
        }
        // Same as doing % 32
        round = (round + 1) & 0b11111;
    }

    finish = true;

    try out.print("Joining\n", .{});

    for (0..WORKER_SIZE) |i| {
        threads[i].join();
    }

    try out.print("Done joining\n", .{});

    const toSort = try alloc.alloc([]const u8, count.keys().len);

    @memcpy(toSort, count.keys());

    try out.print("Start sorting\n", .{});
    std.mem.sort([]const u8, toSort, {}, lessThan);
    try out.print("Finished sorting\n", .{});

    for (toSort) |key| {
        const item = count.get(key).?;
        try out.print("{s}={d:.1}/{d:.1}/{d:.1}; ", .{
            key,
            @as(f64, @floatFromInt(item.min)) / 10.0,
            @as(f64, @floatFromInt(item.mean)) / @as(f64, @floatFromInt(item.count)) / 10.0,
            @as(f64, @floatFromInt(item.max)) / 10.0,
        });
    }
    try out.print("\n", .{});
    // try out.print("\n r > t: {}\n", .{reading_more_faster_than_mux});

    //try out.print("Read {d} lines\n", .{count});
}
